#ifndef DF_OBJECT_H
#define DF_OBJECT_H

#include <glib.h>

#define DF__DECLARATIONS(name, pname)								\
	typedef union _ ## name ## __Type name ## __Type;					\
	typedef struct _ ## name ## __Virtuals name ## __Virtuals;				\
	typedef struct _ ## name ## __Interfaces name ## __Interfaces;				\
	typedef struct _ ## name name;								\
												\
	union _ ## name ## __Type {								\
		guint type;									\
		name ## __TYPE									\
	};											\
	struct _ ## name ## __Virtuals {							\
		name ## __VIRTUALS(name)							\
	};											\
	struct _ ## name ## __Interfaces {							\
		name ## __INTERFACES								\
	};											\
												\
	struct _ ## name {									\
		name ## __Type        _t;							\
		name ## __Virtuals   *_v;							\
		name ## __Interfaces *_i;							\
		name ## __DATA									\
	};											\
												\
	extern pname ## __Virtuals * name ## __ParentVirtuals;					\
	extern pname ## __Interfaces * name ## __ParentInterfaces;				\
												\
	guint name ## __get_type(void);

#define DF__NEW(name)										\
	((name *) DF__Object__create(name ## __get_type()))

#define DF__DEFINITIONS(name, pname, init_virtuals, init_interfaces, init_data)			\
	pname ## __Virtuals * name ## __ParentVirtuals = NULL;					\
	pname ## __Interfaces * name ## __ParentInterfaces = NULL;				\
	guint name ## __get_type (void) {							\
		static guint type = -1;								\
		if (type == -1) {								\
			guint parent_type;							\
			void (*init_v)(name ## __Virtuals *m) = init_virtuals;			\
			void (*init_i)(name ## __Interfaces *i) = init_interfaces;		\
			void (*init_d)(name *self) = init_data;					\
			g_assert(init_v != NULL);						\
			g_assert(init_i != NULL);						\
			g_assert(init_d != NULL);						\
			/* This should be optimized away by the compiler */			\
			if(DF__Object__get_type == name ## __get_type)				\
				parent_type = 0;						\
			else									\
		        	parent_type = pname ## __get_type();				\
			type = DF__Object__init_type(#name,					\
						     parent_type,				\
						     (DF__Object__init)init_v,			\
						     (DF__Object__init)init_i,			\
						     (DF__Object__init_data)init_d,		\
						     sizeof(name),				\
						     sizeof(name ## __Virtuals),		\
						     sizeof(name ## __Interfaces));		\
			name ## __ParentVirtuals = DF__GET_VIRTUALS(pname);			\
			name ## __ParentInterfaces = DF__GET_INTERFACES(pname);			\
		}										\
		return type;									\
	}

#define DF__VIRTUAL_METHOD_PROTOTYPE(return_type, method, args)					\
	return_type (* method ) args;

#define DF__METHOD_SETUP(name, orig_name, return_type, method, args)				\
	static return_type (* const name ## __ ## method) args G_GNUC_UNUSED =			\
		(return_type (* ) args) orig_name ## __ ## method;

#define DF__CALL_VIRTUAL(methods, return_type, method, call_args, no_return)			\
	(return_type) ((methods ## -> ## method) ?						\
		       (methods ## -> ## method) call_args : (return_type)(no_return))

#define DF__GET_VIRTUALS(name)									\
	((name ## __Virtuals *) DF__Object__get_virtuals(name ## __get_type()))
#define DF__GET_INTERFACES(name)								\
	((name ## __Interfaces *) DF__Object__get_interfaces(name ## __get_type()))

#define DF__IS_A(object, name)									\
	(object == NULL ? FALSE : DF__Object__is_type(object->_t.type, name ## __get_type()))

/* Only useful for downcasts, for upcasts use DF__UCAST as that is statically typesafe
 * and this is only runtime typesafe */
#define DF__DCAST(object, name)									\
	((name *)(object == NULL ? NULL :							\
		  (DF__Object__is_type_warn(object->_t.type, name ## __get_type()) ?		\
		   object : NULL)))

#define DF__UCAST(object, name)									\
	((name *)(0 ? (object->_t. ## name ? NULL : NULL) : object))


/*
 * Definition of the DF__Object type
 */

#define DF__Object__TYPE									\
	guint DF__Object;

#define DF__Object__VIRTUALS(type)								\
	void (* destroy)(type *self);								\
	void (* last_unref)(type *self);

#define DF__Object__INTERFACES									\
	gpointer DF__Object__dummy;

#define DF__Object__DATA									\
	int ref_count;

#define DF__Object__METHOD_SETUP(type)								\
	DF__METHOD_SETUP(type, DF__Object, void, ref, (type *self))				\
	DF__METHOD_SETUP(type, DF__Object, void, unref, (type *self))

DF__DECLARATIONS(DF__Object, DF__Object)

/* Actual object methods */
void		DF__Object__ref			(DF__Object *self);
void		DF__Object__unref		(DF__Object *self);

/* Type stuff */

DF__Object *	DF__Object__create		(guint type);
/* This will also emit a warning if type cannot be cast to istype */
gboolean	DF__Object__is_type		(guint type, guint istype);
/* This will also emit a warning if type cannot be cast to istype */
gboolean	DF__Object__is_type_warn	(guint type, guint istype);

typedef void	(*DF__Object__init)		(gpointer data);
typedef void	(*DF__Object__init_data)	(DF__Object *self);

gpointer	DF__Object__get_virtuals	(guint type);
gpointer	DF__Object__get_interfaces	(guint type);

guint 		DF__Object__init_type		(const char			*name,
						 guint				 parent_type,
						 DF__Object__init		 init_virtuals,
						 DF__Object__init		 init_interfaces,
						 DF__Object__init_data		 init_data,
						 guint				 sizeof_object,
						 guint				 sizeof_virtuals,
						 guint				 sizeof_interfaces);

#endif /* DF_OBJECT_H */
